/*
Use a Atmega1284P to flash program into. extended fuse: 0xFC, high fuse: 0x91, low fuse: 0xFD
*/
#define INFO "Radio receiver by Mek version 2.0; RDS decoder: 20060629-1 by Marc Ketel alias DiNo, marc@atoomnet.net."

#include "hd44780.h"
#include "i2cmaster.h"
#include "irmp/irmp.h"
#include "language.h"
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <avr/io.h>
#include <util/delay.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#define _BV(bit) (1 << (bit)) // _BV macro is not defined so define it here
#define BTNF() (bit_is_clear(PINB, 2))
#define BTNR() (bit_is_clear(PINB, 1))
#define BTNFF() (bit_is_clear(PINB, 3))
#define BTNRR() (bit_is_clear(PINB, 0))
#define TUNING_STEP 1 // 1 = 50 kHz
#define TUNING_STEP_FAST 10 // 10 = 500 kHz
#define BASE_FREQ 2496 // 87.50 MHz
#define BASE_FREQ_MHZ 8750 // 87.50 MHz
#define MAX_FREQ 2906 // 108.00 MHz
#define IR_LEARNING_PROGRAM0 1
#define IR_LEARNING_PROGRAM1 2
#define IR_LEARNING_PROGRAM2 3
#define IR_LEARNING_PROGRAM3 4
#define IR_LEARNING_PROGRAM4 5
#define IR_LEARNING_PROGRAM5 6
#define IR_LEARNING_PROGRAM6 7
#define IR_LEARNING_PROGRAM7 8
#define IR_LEARNING_PROGRAM8 9
#define IR_LEARNING_PROGRAM9 10
#define IR_LEARNING_FF 11
#define IR_LEARNING_F 12
#define IR_LEARNING_RR 13
#define IR_LEARNING_R 14
#define IR_LEARNING_MUTE 15
#define IR_LEARNING_PROGRAM 16
#define IR_SET_PROGRAM0 1
#define IR_SET_PROGRAM1 2
#define IR_SET_PROGRAM2 3
#define IR_SET_PROGRAM3 4
#define IR_SET_PROGRAM4 5
#define IR_SET_PROGRAM5 6
#define IR_SET_PROGRAM6 7
#define IR_SET_PROGRAM7 8
#define IR_SET_PROGRAM8 9
#define IR_SET_PROGRAM9 10
#define RDS_PORT PIND
#define RDS_PIN  PD3

volatile unsigned char irLearningMode = 0;
volatile unsigned char programMode = 0;
volatile unsigned int freq;
volatile unsigned int freqMhz;
volatile unsigned int lastFreqSaveCheck = 0;
volatile unsigned int lastFreq = 0;
const unsigned char* LAST_FREQ_ADDRESS = (unsigned char*)98;
static IRMP_DATA irmp_data;
static IRMP_DATA irmp_program0;
static unsigned int freqProgram0;
const unsigned char* EEPROM_ADDRESS_PROGRAM0 = (unsigned char*)100;
const unsigned char* EEPROM_ADDRESS_PROGRAM0_FREQ = (unsigned char*)106; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program1;
static unsigned int freqProgram1;
const unsigned char* EEPROM_ADDRESS_PROGRAM1 = (unsigned char*)110;
const unsigned char* EEPROM_ADDRESS_PROGRAM1_FREQ = (unsigned char*)116; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program2;
static unsigned int freqProgram2;
const unsigned char* EEPROM_ADDRESS_PROGRAM2 = (unsigned char*)120;
const unsigned char* EEPROM_ADDRESS_PROGRAM2_FREQ = (unsigned char*)126; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program3;
static unsigned int freqProgram3;
const unsigned char* EEPROM_ADDRESS_PROGRAM3 = (unsigned char*)130;
const unsigned char* EEPROM_ADDRESS_PROGRAM3_FREQ = (unsigned char*)136; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program4;
static unsigned int freqProgram4;
const unsigned char* EEPROM_ADDRESS_PROGRAM4 = (unsigned char*)140;
const unsigned char* EEPROM_ADDRESS_PROGRAM4_FREQ = (unsigned char*)146; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program5;
static unsigned int freqProgram5;
const unsigned char* EEPROM_ADDRESS_PROGRAM5 = (unsigned char*)150;
const unsigned char* EEPROM_ADDRESS_PROGRAM5_FREQ = (unsigned char*)156; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program6;
static unsigned int freqProgram6;
const unsigned char* EEPROM_ADDRESS_PROGRAM6 = (unsigned char*)160;
const unsigned char* EEPROM_ADDRESS_PROGRAM6_FREQ = (unsigned char*)166; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program7;
static unsigned int freqProgram7;
const unsigned char* EEPROM_ADDRESS_PROGRAM7 = (unsigned char*)170;
const unsigned char* EEPROM_ADDRESS_PROGRAM7_FREQ = (unsigned char*)176; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program8;
static unsigned int freqProgram8;
const unsigned char* EEPROM_ADDRESS_PROGRAM8 = (unsigned char*)180;
const unsigned char* EEPROM_ADDRESS_PROGRAM8_FREQ = (unsigned char*)186; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program9;
static unsigned int freqProgram9;
const unsigned char* EEPROM_ADDRESS_PROGRAM9 = (unsigned char*)190;
const unsigned char* EEPROM_ADDRESS_PROGRAM9_FREQ = (unsigned char*)196; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_rr;
const unsigned char* EEPROM_ADDRESS_RR = (unsigned char*)200; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_r;
const unsigned char* EEPROM_ADDRESS_R = (unsigned char*)206; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_ff;
const unsigned char* EEPROM_ADDRESS_FF = (unsigned char*)212; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_f;
const unsigned char* EEPROM_ADDRESS_F = (unsigned char*)218; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_mute;
const unsigned char* EEPROM_ADDRESS_MUTE = (unsigned char*)224; // 6 is sizeof(IRMP_DATA)
static IRMP_DATA irmp_program;
const unsigned char* EEPROM_ADDRESS_PROGRAM = (unsigned char*)230; // 6 is sizeof(IRMP_DATA)

#define max(a,b) \
({ __typeof__ (a) _a = (a); \
	__typeof__ (b) _b = (b); \
_a > _b ? _a : _b; })

FILE* Uart;

//used for stdout
int uartSend(char data, FILE* stream) {
	//wait for completion of previous send
	loop_until_bit_is_set(UCSR0A, UDRE0);

	//send the data
	UDR0 = data;
	return 0;
}

//used for stdin
int uartRecv(FILE* stream) {
	loop_until_bit_is_set(UCSR0A, RXC0);
	return UDR0;
}

#define eventTimerOverflow      _BV(0)
#define eventUsart0RxDInterrupt _BV(1)
#define eventGroupComplete      _BV(2)
#define eventBitstreamSnapshot  _BV(3)
#define eventSyncLost           _BV(4)

volatile unsigned char event = 0;
volatile unsigned char timerOverflowCount = 0;
volatile unsigned char usart0RxDByte = 0;
volatile unsigned char bitCounter = 0;
volatile unsigned char bitCounterSnapshot = 0;

volatile unsigned char synchronized = 0;
volatile unsigned char syncQuality = 0; //0-31, 0=no usable signal, 15=medium quality, 31=perfect!
volatile unsigned char syncCounter = 0;

volatile unsigned char block = 0;
volatile unsigned long bitstream = 0;
volatile unsigned long bitstreamData = 0;
volatile unsigned int block1 = 0;
volatile unsigned int block2 = 0;
volatile unsigned int block3 = 0;
volatile unsigned int block4 = 0;

volatile char radioText[64];
volatile int isRDS = 0;
volatile int radioOn = 0;
unsigned char programmeServiceName[8];
unsigned char oldProgrammeServiceName[8];

volatile char radioTextDisplayStart = 0;
volatile int radioTextMaxIndex = -1;
volatile char radioTextZeroDisplayCount = 0;
volatile char radioTextEndDisplayCount = 0;
volatile char timer1compaCount = 0;
volatile unsigned int timer2overflowCount = 0;
volatile unsigned int isNormalMode = 1; // indicates whether radio is in normal mode, displaying RDS info... or in some special mode like IR code learning or program saving

void showPs()
{
	// compare old to new one to avoid flickering
	if (isNormalMode == 1 && memcmp(&programmeServiceName[0], &oldProgrammeServiceName[0], sizeof(programmeServiceName)) != 0)
	{
		lcd_home();
		lcd_puts("        "); // PS is max 8 chars, to the right frequency is displayed
		lcd_home();
		for(unsigned int i = 0; i < sizeof(programmeServiceName) && i <= 7; i++) {
			lcd_putc((programmeServiceName[i] < 32 || programmeServiceName[i] > 126) ? ' ' : programmeServiceName[i]); // don't display gibberish if it is there
		}
	
		memcpy(oldProgrammeServiceName, &programmeServiceName[0], sizeof(programmeServiceName));
	}
}

void showFreq()
{
	if (isNormalMode == 1)
	{
		lcd_goto(0x09);
		
		if (freqMhz < 10000)
		{
			lcd_putc(' ');
			lcd_goto(0x0a);
			lcd_putc('0' + freqMhz / 1000);
			lcd_goto(0x0b);
			lcd_putc('0' + (freqMhz % 1000) / 100);
			lcd_goto(0x0c);
			lcd_putc(L_DECIMAL_DOT);
			lcd_goto(0x0d);
			lcd_putc('0' + (freqMhz % 100) / 10);
			lcd_goto(0x0e);
			lcd_putc('0' + freqMhz % 10);
			lcd_goto(0x0f);
			lcd_putc('M');
		} else {
			lcd_putc('1');
			lcd_goto(0x0a);
			lcd_putc('0' + (freqMhz % 10000) / 1000);
			lcd_goto(0x0b);
			lcd_putc('0' + (freqMhz % 1000) / 100);
			lcd_goto(0x0c);
			lcd_putc(L_DECIMAL_DOT);
			lcd_goto(0x0d);
			lcd_putc('0' + (freqMhz % 100) / 10);
			lcd_goto(0x0e);
			lcd_putc('0' + freqMhz % 10);
			lcd_goto(0x0f);
			lcd_putc('M');
		}
	}
}

void showNoRds()
{
	if (isNormalMode == 1)
	{
		lcd_home();
		lcd_puts("   FM   ");
		lcd_goto(0x40);
		lcd_puts(L_NO_RDS_SIGNAL);
	}
}

void setIsNormalMode(unsigned int value)
{
	isNormalMode = value;
	
	if (value == 1)
	{
		// reflect normal mode settings: display RDS info
		lcd_clrscr();
		showFreq();
		
		if (isRDS == 1)
		{
			memset(&oldProgrammeServiceName, ' ', sizeof(oldProgrammeServiceName));
			showPs();
		}
		else
		{
			showNoRds();
		}
	}
}

static void learnIrCode(const unsigned char* eepromAddress, IRMP_DATA* irmpDataToSet, IRMP_DATA* irmpDataVariable)
{
	eeprom_write_block(irmpDataToSet, (unsigned char *)eepromAddress, sizeof(*irmpDataToSet));
	memcpy(irmpDataVariable, irmpDataToSet, sizeof(*irmpDataToSet));
}

static void setFreq()
{
	unsigned char dividerH = freq >> 8;
	unsigned char dividerL = freq;
	
	freqMhz = BASE_FREQ_MHZ + ((freq - BASE_FREQ) * 5); // 1 in 'freq' == 50 kHz
	
	i2c_start_wait(0xc2 + I2C_WRITE); // set device address and write mode
	i2c_write(dividerH); // program divider byte 1 (msb)
	i2c_write(dividerL); // program divider byte 2 (lsb)
	i2c_write(0b10001000); // control byte
	i2c_write(0b00000001); // switch to low band
	i2c_stop();

	if (radioOn == 1)
	{
		showFreq();
	}
	
	lastFreqSaveCheck = 0; // invoke last frequency save to eeprom after a few seconds to avoid eeprom wear
}

static void fastBack_rr()
{
	if (freq == BASE_FREQ)
	{
		freq = MAX_FREQ;
	} else if (freq - TUNING_STEP_FAST < BASE_FREQ) {
		freq = BASE_FREQ;
	} else {
		freq -= TUNING_STEP_FAST;
	}

	setFreq();
}

static void back_r()
{
	if (freq == BASE_FREQ)
	{
		freq = MAX_FREQ;
	} else if (freq - TUNING_STEP < BASE_FREQ) {
		freq = BASE_FREQ;
	} else {
		freq -= TUNING_STEP;
	}

	setFreq();
}

static void forward_f()
{
	if (freq == MAX_FREQ)
	{
		freq = BASE_FREQ;
	} else if (freq + TUNING_STEP > MAX_FREQ) {
		freq = MAX_FREQ;
	} else {
		freq += TUNING_STEP;
	}

	setFreq();
}

static void fastForward_ff()
{
	if (freq == MAX_FREQ)
	{
		freq = BASE_FREQ;
	} else if (freq + TUNING_STEP_FAST > MAX_FREQ) {
		freq = MAX_FREQ;
	} else {
		freq += TUNING_STEP_FAST;
	}

	setFreq();
}

//ISR is executed when there is a new rds bit to read
ISR(INT0_vect) {
	cli();
	unsigned int syndrome = 0;
	
	//Copy global vars to local vars for performance
	unsigned long bitstreamLocal = bitstream;
	unsigned char blockLocal = block;
	
	//shift rds bit in on right side
	bitstreamLocal *= 2;
	if (RDS_PORT & _BV(RDS_PIN))
	bitstreamLocal++;
	
	bitCounter++;
	
	//collect data for raw bitstream snapshots
	bitCounterSnapshot++;
	if (bitCounterSnapshot == 26) {
		bitstreamData = (bitstreamLocal & 0x3FFFFFF);
		bitCounterSnapshot = 0;
		
		event |= eventBitstreamSnapshot;
	}
	
	//when we have 26 bits or are not synchronized to the stream
	//check the CRC and maybe store a rds block
	if (bitCounter == 26 || !synchronized) {
		
		bitstreamLocal &= 0x3FFFFFF; //we only need 26 bits
		
		syndrome = (bitstreamLocal / 0x10000); //bits 16-31 (2^16)
		
		if (bitstreamLocal & _BV(0))
		syndrome ^= 0x031b;
		
		if (bitstreamLocal & _BV(1))
		syndrome ^= 0x038f;
		
		if (bitstreamLocal & _BV(2))
		syndrome ^= 0x02a7;
		
		if (bitstreamLocal & _BV(3))
		syndrome ^= 0x00f7;
		
		if (bitstreamLocal & _BV(4))
		syndrome ^= 0x01ee;
		
		if (bitstreamLocal & _BV(5))
		syndrome ^= 0x03dc;
		
		if (bitstreamLocal & _BV(6))
		syndrome ^= 0x0201;
		
		if (bitstreamLocal & _BV(7))
		syndrome ^= 0x01bb;
		
		if (bitstreamLocal & _BV(8))
		syndrome ^= 0x0376;
		
		if (bitstreamLocal & _BV(9))
		syndrome ^= 0x0355;
		
		if (bitstreamLocal & _BV(10))
		syndrome ^= 0x0313;
		
		if (bitstreamLocal & _BV(11))
		syndrome ^= 0x039f;
		
		if (bitstreamLocal & _BV(12))
		syndrome ^= 0x0287;
		
		if (bitstreamLocal & _BV(13))
		syndrome ^= 0x00b7;
		
		if (bitstreamLocal & _BV(14))
		syndrome ^= 0x016e;

		if (bitstreamLocal & 0b1000000000000000) // _BV(15) does not work!!!
		syndrome ^= 0x02dc;

		//Block A?
		if (blockLocal == 0 && syndrome == 0x03d8) {
			block1 = bitstreamLocal / 0x400; //bits 10-25 (2^10)
			synchronized = 1;
			blockLocal++;
		} else
		//Block B?
		if (blockLocal == 1 && syndrome == 0x03d4) {
			block2 = bitstreamLocal / 0x400; //bits 10-25 (2^10)
			synchronized = 1;
			blockLocal++;
		} else
		//Block C type A?
		if (blockLocal == 2 && !(block2 & _BV(11)) && syndrome == 0x025c) {
			block3 = bitstreamLocal / 0x400; //bits 10-25 (2^10)
			synchronized = 1;
			blockLocal++;
		} else
		//Block C type B?
		if (blockLocal == 2 && (block2 & _BV(11)) && syndrome == 0x03cc) {
			block3 = bitstreamLocal / 0x400; //bits 10-25 (2^10)
			synchronized = 1;
			blockLocal++;
		} else
		//Block D?
		if (blockLocal == 3 && syndrome == 0x0258) {
			block4 = bitstreamLocal / 0x400; //bits 10-25 (2^10)
			synchronized = 1;
			blockLocal = 0;
			//we have a complete group!
			event |= eventGroupComplete;
		} else {
			//sync lost..
			synchronized = 0;
			blockLocal = 0;
			event |= eventSyncLost;
		}
		
		bitCounter = 0;
	}
	
	if (event & eventGroupComplete) {
		//PORTB |= _BV(PB5);
		
		if (syncQuality < 31)
		syncQuality++;
		
	} else if (!synchronized) {
		//PORTB &= ~_BV(PB5);
		
		syncCounter++;
		if (syncQuality > 0 && syncCounter == 26) {
			syncQuality--;
			syncCounter = 0;
		}
		
	}
	
	//if (syncQuality >= 15)
	//PORTC |= _BV(PC2);
	//else
	//PORTC &= ~_BV(PC2);
	
	//Store local vars into global vars to remember state
	bitstream = bitstreamLocal;
	block = blockLocal;
	sei();
}

//timer0 overflowed
ISR(TIMER0_OVF_vect) {
	cli();
	timerOverflowCount++;
	if (timerOverflowCount > 4) {
		event |= eventTimerOverflow;
		timerOverflowCount = 0;
	}
	sei();
}

//timer2 overflowed, it overflows 20 times in second
ISR(TIMER2_OVF_vect) {
	timer2overflowCount++;
	const unsigned char initialThreshold = 30;
	
	// turn on signal input (late to avoid short noise from FM radio at the beginning)
	if (radioOn == 0 && timer2overflowCount == initialThreshold)
	{
		PORTA |= _BV(PA7);
		radioOn = 1;
		showFreq();
		showNoRds();
	}
	
	if (isNormalMode == 1 && isRDS == 1 && timer2overflowCount == initialThreshold)
	{
		if (radioOn == 1) {
			showPs();
		}

		timer2overflowCount = 0;
		//char subbuff[17];
		//memcpy(subbuff, &radioText[0], max(radioTextMaxIndex, 16)); // first 16 chars
		//subbuff[16] = '\0';
		//lcd_puts(subbuff);
		if (radioTextMaxIndex <= 15)
		{
			lcd_goto(0x40);
			lcd_puts("                ");
			lcd_goto(0x40);
			for (int i = 0; i <= radioTextMaxIndex; i++)
			{
				lcd_putc(radioText[i]);
			}
		}
		else
		{
			lcd_goto(0x40);
			lcd_puts("                ");
			lcd_goto(0x40);
			for (int i = radioTextDisplayStart; i <= radioTextDisplayStart + 15 && i <= radioTextMaxIndex; i++)
			{
				lcd_putc(radioText[i]);
			}
			
			if (radioTextDisplayStart + 16 > radioTextMaxIndex)
			{
				if (radioTextEndDisplayCount > 5)
				{
					// move to the beginning
					radioTextDisplayStart = 0;
					radioTextZeroDisplayCount = 0;
					radioTextEndDisplayCount = 0;
				}
				else
				{
					// stay at the end
					radioTextEndDisplayCount++;
				}
			} else {
				if (radioTextZeroDisplayCount > 5)
				{
					// scroll one position right
					radioTextDisplayStart++;
				}
				else
				{
					// stay at the beginning
					radioTextZeroDisplayCount++;
				}
			}
		}
	}
	
	if (timer2overflowCount == initialThreshold)
	{
		timer2overflowCount = 0;
	}
	
	if (lastFreqSaveCheck == 100) // after few seconds when no tuning was made, save current frequency to eeprom, this is to avoid its wear
	{
		if (freq != lastFreq) // this prevents overwriting with the same value if only turned on without tuning yet
		{
			eeprom_update_byte((unsigned char*)LAST_FREQ_ADDRESS, freq & 0xFF);
			eeprom_update_byte((unsigned char*)LAST_FREQ_ADDRESS + 1, freq >> 8);
			lastFreq = freq;
		}
	} else {
		lastFreqSaveCheck++;
	}
	
	// disable buttons if radio is not on yet, or we are in some special mode
	if (isNormalMode == 1 && radioOn == 1)
	{
		if (BTNF() && BTNR())
		{
			// enter IR code learning mode
			setIsNormalMode(0);
			lcd_clrscr();
			lcd_home();
			lcd_puts(L_REMOTE_LEARNING);
			lcd_goto(0x40);
			lcd_puts(L_LEARNING_PROGRAM1);
			irLearningMode = IR_LEARNING_PROGRAM0;
		}
		else if (BTNF())
		{
			forward_f();
			wdt_reset();
			_delay_ms(500);
		}
		else if (BTNR())
		{
			back_r();
			wdt_reset();
			_delay_ms(500);
		}
		else if (BTNFF())
		{
			fastForward_ff();
			wdt_reset();
			_delay_ms(500);
		}
		else if (BTNRR())
		{
			fastBack_rr();
			wdt_reset();
			_delay_ms(500);
		}
	}
		
	if (irmp_get_data(&irmp_data) && !(irmp_data.flags & IRMP_FLAG_REPETITION)) // repetition is not supported because it's unwantedly triggered even on single button press
	{
		/* ir signal decoded, do something here... */
		/* irmp_data.protocol is the protocol, see irmp.h */
		/* irmp_data.address is the address/manufacturer code of ir sender */
		/* irmp_data.command is the command code */
		if (irLearningMode != 0)
		{
			switch (irLearningMode)
			{
				case IR_LEARNING_PROGRAM0:
				learnIrCode(EEPROM_ADDRESS_PROGRAM0, &irmp_data, &irmp_program0);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_PROGRAM2);
				irLearningMode = IR_LEARNING_PROGRAM1;
				break;
				case IR_LEARNING_PROGRAM1:
				learnIrCode(EEPROM_ADDRESS_PROGRAM1, &irmp_data, &irmp_program1);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_PROGRAM3);
				irLearningMode = IR_LEARNING_PROGRAM2;
				break;
				case IR_LEARNING_PROGRAM2:
				learnIrCode(EEPROM_ADDRESS_PROGRAM2, &irmp_data, &irmp_program2);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_PROGRAM4);
				irLearningMode = IR_LEARNING_PROGRAM3;
				break;
				case IR_LEARNING_PROGRAM3:
				learnIrCode(EEPROM_ADDRESS_PROGRAM3, &irmp_data, &irmp_program3);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_PROGRAM5);
				irLearningMode = IR_LEARNING_PROGRAM4;
				break;
				case IR_LEARNING_PROGRAM4:
				learnIrCode(EEPROM_ADDRESS_PROGRAM4, &irmp_data, &irmp_program4);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_PROGRAM6);
				irLearningMode = IR_LEARNING_PROGRAM5;
				break;
				case IR_LEARNING_PROGRAM5:
				learnIrCode(EEPROM_ADDRESS_PROGRAM5, &irmp_data, &irmp_program5);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_PROGRAM7);
				irLearningMode = IR_LEARNING_PROGRAM6;
				break;
				case IR_LEARNING_PROGRAM6:
				learnIrCode(EEPROM_ADDRESS_PROGRAM6, &irmp_data, &irmp_program6);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_PROGRAM8);
				irLearningMode = IR_LEARNING_PROGRAM7;
				break;
				case IR_LEARNING_PROGRAM7:
				learnIrCode(EEPROM_ADDRESS_PROGRAM7, &irmp_data, &irmp_program7);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_PROGRAM9);
				irLearningMode = IR_LEARNING_PROGRAM8;
				break;
				case IR_LEARNING_PROGRAM8:
				learnIrCode(EEPROM_ADDRESS_PROGRAM8, &irmp_data, &irmp_program8);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_PROGRAM10);
				irLearningMode = IR_LEARNING_PROGRAM9;
				break;
				case IR_LEARNING_PROGRAM9:
				learnIrCode(EEPROM_ADDRESS_PROGRAM9, &irmp_data, &irmp_program9);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_RR);
				irLearningMode = IR_LEARNING_RR;
				break;
				case IR_LEARNING_RR:
				learnIrCode(EEPROM_ADDRESS_RR, &irmp_data, &irmp_rr);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_R);
				irLearningMode = IR_LEARNING_R;
				break;
				case IR_LEARNING_R:
				learnIrCode(EEPROM_ADDRESS_R, &irmp_data, &irmp_r);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_FF);
				irLearningMode = IR_LEARNING_FF;
				break;
				case IR_LEARNING_FF:
				learnIrCode(EEPROM_ADDRESS_FF, &irmp_data, &irmp_ff);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_F);
				irLearningMode = IR_LEARNING_F;
				break;
				case IR_LEARNING_F:
				learnIrCode(EEPROM_ADDRESS_F, &irmp_data, &irmp_f);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_MUTE);
				irLearningMode = IR_LEARNING_MUTE;
				break;
				case IR_LEARNING_MUTE:
				learnIrCode(EEPROM_ADDRESS_MUTE, &irmp_data, &irmp_mute);
				lcd_goto(0x40);
				lcd_puts(L_LEARNING_PROGRAM);
				irLearningMode = IR_LEARNING_PROGRAM;
				break;
				case IR_LEARNING_PROGRAM:
				learnIrCode(EEPROM_ADDRESS_PROGRAM, &irmp_data, &irmp_program);
				irLearningMode = 0;
				setIsNormalMode(1);
				break;
			}
		}
		else if (programMode == 1)
		{
			unsigned char wasRecognized = 0;
			
			// user pressed button to save a program into, let's check if it is valid Program button, or the Program button was pressed again to cancel
			if (memcmp(&irmp_data, &irmp_program0, sizeof(irmp_data) - 1) == 0)
			{
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM0_FREQ, freq & 0xFF);
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM0_FREQ + 1, freq >> 8);
				freqProgram0 = freq;
				wasRecognized = 1;
			}
			else if (memcmp(&irmp_data, &irmp_program1, sizeof(irmp_data) - 1) == 0)
			{
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM1_FREQ, freq & 0xFF);
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM1_FREQ + 1, freq >> 8);
				freqProgram1 = freq;
				wasRecognized = 1;
			}
			else if (memcmp(&irmp_data, &irmp_program2, sizeof(irmp_data) - 1) == 0)
			{
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM2_FREQ, freq & 0xFF);
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM2_FREQ + 1, freq >> 8);
				freqProgram2 = freq;
				wasRecognized = 1;
			}
			else if (memcmp(&irmp_data, &irmp_program3, sizeof(irmp_data) - 1) == 0)
			{
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM3_FREQ, freq & 0xFF);
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM3_FREQ + 1, freq >> 8);
				freqProgram3 = freq;
				wasRecognized = 1;
			}
			else if (memcmp(&irmp_data, &irmp_program4, sizeof(irmp_data) - 1) == 0)
			{
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM4_FREQ, freq & 0xFF);
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM4_FREQ + 1, freq >> 8);
				freqProgram4 = freq;
				wasRecognized = 1;
			}
			else if (memcmp(&irmp_data, &irmp_program5, sizeof(irmp_data) - 1) == 0)
			{
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM5_FREQ, freq & 0xFF);
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM5_FREQ + 1, freq >> 8);
				freqProgram5 = freq;
				wasRecognized = 1;
			}
			else if (memcmp(&irmp_data, &irmp_program6, sizeof(irmp_data) - 1) == 0)
			{
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM6_FREQ, freq & 0xFF);
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM6_FREQ + 1, freq >> 8);
				freqProgram6 = freq;
				wasRecognized = 1;
			}
			else if (memcmp(&irmp_data, &irmp_program7, sizeof(irmp_data) - 1) == 0)
			{
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM7_FREQ, freq & 0xFF);
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM7_FREQ + 1, freq >> 8);
				freqProgram7 = freq;
				wasRecognized = 1;
			}
			else if (memcmp(&irmp_data, &irmp_program8, sizeof(irmp_data) - 1) == 0)
			{
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM8_FREQ, freq & 0xFF);
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM8_FREQ + 1, freq >> 8);
				freqProgram8 = freq;
				wasRecognized = 1;
			}
			else if (memcmp(&irmp_data, &irmp_program9, sizeof(irmp_data) - 1) == 0)
			{
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM9_FREQ, freq & 0xFF);
				eeprom_update_byte((unsigned char*)EEPROM_ADDRESS_PROGRAM9_FREQ + 1, freq >> 8);
				freqProgram9 = freq;
				wasRecognized = 1;
			}
			else if (memcmp(&irmp_data, &irmp_program, sizeof(irmp_data) - 1) == 0)
			{
				wasRecognized = 1;
			}
			
			if (wasRecognized == 1)
			{
				programMode = 0;
				setIsNormalMode(1);
			}
		}
		else
		{
			// compare trained code with received code
			// below we compare sizeof - 1 because we want to exclude 'flags' property (it indicates repetition)
			if ((freqProgram0 >= BASE_FREQ && freqProgram0 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program0, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram0;
				setFreq();
			}
			else if ((freqProgram1 >= BASE_FREQ && freqProgram1 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program1, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram1;
				setFreq();
			}
			else if ((freqProgram2 >= BASE_FREQ && freqProgram2 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program2, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram2;
				setFreq();
			}
			else if ((freqProgram3 >= BASE_FREQ && freqProgram3 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program3, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram3;
				setFreq();
			}
			else if ((freqProgram4 >= BASE_FREQ && freqProgram4 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program4, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram4;
				setFreq();
			}
			else if ((freqProgram5 >= BASE_FREQ && freqProgram5 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program5, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram5;
				setFreq();
			}
			else if ((freqProgram6 >= BASE_FREQ && freqProgram6 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program6, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram6;
				setFreq();
			}
			else if ((freqProgram7 >= BASE_FREQ && freqProgram7 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program7, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram7;
				setFreq();
			}
			else if ((freqProgram8 >= BASE_FREQ && freqProgram8 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program8, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram8;
				setFreq();
			}
			else if ((freqProgram9 >= BASE_FREQ && freqProgram9 <= MAX_FREQ) && memcmp(&irmp_data, &irmp_program9, sizeof(irmp_data) - 1) == 0)
			{
				// switch to saved frequency
				freq = freqProgram9;
				setFreq();
			}
			else if (memcmp(&irmp_data, &irmp_ff, sizeof(irmp_data) - 1) == 0)
			{
				fastForward_ff();
			}
			else if (memcmp(&irmp_data, &irmp_f, sizeof(irmp_data) - 1) == 0)
			{
				forward_f();
			}
			else if (memcmp(&irmp_data, &irmp_rr, sizeof(irmp_data) - 1) == 0)
			{
				fastBack_rr();
			}
			else if (memcmp(&irmp_data, &irmp_r, sizeof(irmp_data) - 1) == 0)
			{
				back_r();
			}
			else if (memcmp(&irmp_data, &irmp_mute, sizeof(irmp_data) - 1) == 0)
			{
				PORTA ^= _BV(PA7);
			}
			else if (memcmp(&irmp_data, &irmp_program, sizeof(irmp_program) - 1) == 0)
			{
				setIsNormalMode(0);
				lcd_clrscr();
				lcd_home();
				lcd_puts(L_SAVE_PROGRAM);
				lcd_goto(0x40);
				lcd_puts(L_PRESS_REMOTE_BTN);
				
				programMode = 1;
			}
		}
	}
}

ISR(USART0_RX_vect) {
	cli();
	usart0RxDByte = UDR0;
	event |= eventUsart0RxDInterrupt;
	sei();
}

ISR(TIMER1_COMPA_vect)
{
	(void)irmp_ISR();
}

void displayInfo(void) {
	printf_P(PSTR("INFO: 0x01, "));
	printf_P(PSTR(INFO));
	printf_P(PSTR("\r\n"));
}

static void irmp_timer_init()
{
	#if defined (__AVR_ATtiny45__) || defined (__AVR_ATtiny85__)                // ATtiny45 / ATtiny85:

	#if F_CPU >= 16000000L
	OCR1C   =  (F_CPU / F_INTERRUPTS / 8) - 1;                              // compare value: 1/15000 of CPU frequency, presc = 8
	TCCR1   = (1 << CTC1) | (1 << CS12);                                    // switch CTC Mode on, set prescaler to 8
	#else
	OCR1C   =  (F_CPU / F_INTERRUPTS / 4) - 1;                              // compare value: 1/15000 of CPU frequency, presc = 4
	TCCR1   = (1 << CTC1) | (1 << CS11) | (1 << CS10);                      // switch CTC Mode on, set prescaler to 4
	#endif

	#else                                                                       // ATmegaXX:
	OCR1A   =  (F_CPU / F_INTERRUPTS) - 1;                                  // compare value: 1/15000 of CPU frequency
	TCCR1B  = (1 << WGM12) | (1 << CS10);                                   // switch CTC Mode on, set prescaler to 1
	#endif

	#ifdef TIMSK1
	TIMSK1  = 1 << OCIE1A;                                                  // OCIE1A: Interrupt by timer compare
	#else
	TIMSK   = 1 << OCIE1A;                                                  // OCIE1A: Interrupt by timer compare
	#endif
}

int main(void)
{
	wdt_reset(); // reset immediately in case of a previous system reset
	wdt_enable(WDTO_1S); // lets enable watchdog with 1 second timeout
	
	unsigned int  programmeIdentificationCode = 0;
	unsigned char groupType = 0;
	unsigned char groupVersion = 0;
	unsigned char trafficProgrammeIdentificationCode = 0;
	unsigned char programmeTypeCode = 0;
	unsigned char trafficAnnouncementCode = 0;
	unsigned char musicSpeechSwitchScode = 0;
	unsigned char group0CodeBits = 0;
	unsigned char group0CodeBitsSteadyCount = 0;
	unsigned char programmeServiceNameNew[8];
	unsigned char decoderIdentificationControlCode = 0;
	unsigned char decoderIdentificationControlCodeNew = 0;
	unsigned char alternativeFrequencyCodes[27];
	unsigned char syncMessage = 0;
	unsigned char displayRaw = 0;
	unsigned char displayBitstream = 0;
	unsigned char linkageActuator = 0;
	unsigned char extendedCountryCode = 0;
	unsigned char textSegmentAddress = 0;
	unsigned char textSegmentAddressPrevious = 0;
	unsigned char textVersion = 0;
	unsigned char textVersionPrevious = 0;
	unsigned char radioTextPrevious[64];
	unsigned char textSegmentAddress0Seen = 0;
	unsigned int  modifiedJulianDay = 0;
	unsigned int  utcYear = 0;
	unsigned char utcMonth = 0;
	unsigned char utcDay = 0;
	signed   char localHours = 0;
	unsigned char utcHours = 0;
	signed   char localMinutes = 0;
	unsigned char utcMinutes = 0;
	unsigned char utcMinutesPrevious = 0xFF;
	unsigned char localSign = 0;
	unsigned int  localTimeOffset = 0;
  
	//general purpose vars
	unsigned int  m;
	unsigned char h;
	unsigned char i;
	unsigned char j;
	
	//CONFIGURE: 8 BIT DATA, 1 STOP BIT, NO PARITY
	UCSR0C = _BV(UCSZ01) | _BV(UCSZ00);
		
	//4.332Mhz crystal
	//baudrate 38K4
	UBRR0 = 6;
  
	//UBBR0 should be ((F_CPU/(USART_BAUDRATE*16UL))-1)
  
	// 4mhz, 9600
	//UBRR0 = 25;
  
	//enable RxD interrupt, RxD, TxD
	UCSR0B |= _BV(RXCIE0) | _BV(RXEN0) | _BV(TXEN0);
	
	//UART is stdout and stdin;
	Uart = fdevopen(uartSend, uartRecv);

	//enable int0
	EIMSK = _BV(INT0);
  
	//INT0 raising edge
	EICRA |= _BV(ISC01) | _BV(ISC00);
	
	// outputs
	DDRA |= _BV(PA7);
	
	// inputs
	DDRD &= ~_BV(PD3);
	DDRD &= ~_BV(PD7);
	DDRB &= ~_BV(PB0);
	DDRB &= ~_BV(PB1);
	DDRB &= ~_BV(PB2);
	DDRB &= ~_BV(PB3);
	DDRB &= ~_BV(PB4);
	
	//timer0 - required by RDS - prescaler clk/1024
	TCCR0B |= _BV(CS02) | _BV(CS00);
	//enable overflow interrupt
	TIMSK0 |= _BV(TOIE0);
	
	//timer1 - used by IRMP, is configured in irmp_timer_init()
	
	//timer2 - prescaler clk/1024
	TCCR2B |= _BV(CS22) | _BV(CS21) | _BV(CS20);
	//enable overflow interrupt
	TIMSK2 |= _BV(TOIE2);
	
	displayInfo();
  
	lcd_init();
	lcd_clrscr();
	lcd_command(_BV(LCD_DISPLAYMODE) | _BV(LCD_DISPLAYMODE_ON)); // display begins in OFF state, turn it ON
	
	// setup custom characters for LCD
	// enter CGRAM mode (after each character, CGRAM pointer is moved to next position)
	lcd_command(0x40);
	// nothing, address = 0 (is unusable because it is treated as end of string by lcd_puts function)
	lcd_putc(0b00001010);
	lcd_putc(0b00000100);
	lcd_putc(0b00001111);
	lcd_putc(0b00010000);
	lcd_putc(0b00010000);
	lcd_putc(0b00010000);
	lcd_putc(0b00001111);
	lcd_putc(0);
	// , address = 1
	lcd_putc(0b00001010);
	lcd_putc(0b00000100);
	lcd_putc(0b00001111);
	lcd_putc(0b00010000);
	lcd_putc(0b00010000);
	lcd_putc(0b00010000);
	lcd_putc(0b00001111);
	lcd_putc(0);
	// , address = 2
	lcd_putc(0b00001101);
	lcd_putc(0b00000101);
	lcd_putc(0b00000110);
	lcd_putc(0b00000100);
	lcd_putc(0b00000100);
	lcd_putc(0b00000100);
	lcd_putc(0b00001110);
	lcd_putc(0);
	// exit CGRAM mode
	lcd_command(0x80);
	
	lcd_clrscr();
	lcd_home();
	lcd_puts("FM Radio by Mek ");
	lcd_goto(0x40);
	lcd_puts(" www.mekweb.eu  ");
	
	_delay_ms(300);
	
	unsigned int r;
	eeprom_read_block(&r, (unsigned char *)LAST_FREQ_ADDRESS, sizeof(r));
	lastFreq = r;
	freq = (lastFreq >= BASE_FREQ && lastFreq <= MAX_FREQ) ? lastFreq : BASE_FREQ;
	
	i2c_init();
	memset(&irmp_data, 0x00, sizeof(irmp_data));
	irmp_init();
	irmp_timer_init();
	
	// i2c scan
	//char addr[2];
	//for (unsigned char i = 0; i <= 255; i++) {
		//sprintf(addr, "%02X", i);
		//lcd_goto(0x40);  // sem dat conditional breakpoint, spomali to beh programu. condition: i == 0xc2
		//lcd_puts(addr);
		//if (i2c_start(i) == 0) {
			//lcd_goto(0x09); // sem dat breakpoint
			//lcd_puts(addr);
		//}
		//for (unsigned long ii = 0; ii <= 0xffff; ii++);
	//}
	
	sei(); // enable interrupts
	
	setFreq();
	
	eeprom_read_block(&irmp_program0, (unsigned char *)EEPROM_ADDRESS_PROGRAM0, sizeof(irmp_program0));
	eeprom_read_block(&freqProgram0, (unsigned char *)EEPROM_ADDRESS_PROGRAM0_FREQ, sizeof(freqProgram0));
	eeprom_read_block(&irmp_program1, (unsigned char *)EEPROM_ADDRESS_PROGRAM1, sizeof(irmp_program1));
	eeprom_read_block(&freqProgram1, (unsigned char *)EEPROM_ADDRESS_PROGRAM1_FREQ, sizeof(freqProgram1));
	eeprom_read_block(&irmp_program2, (unsigned char *)EEPROM_ADDRESS_PROGRAM2, sizeof(irmp_program2));
	eeprom_read_block(&freqProgram2, (unsigned char *)EEPROM_ADDRESS_PROGRAM2_FREQ, sizeof(freqProgram2));
	eeprom_read_block(&irmp_program3, (unsigned char *)EEPROM_ADDRESS_PROGRAM3, sizeof(irmp_program3));
	eeprom_read_block(&freqProgram3, (unsigned char *)EEPROM_ADDRESS_PROGRAM3_FREQ, sizeof(freqProgram3));
	eeprom_read_block(&irmp_program4, (unsigned char *)EEPROM_ADDRESS_PROGRAM4, sizeof(irmp_program4));
	eeprom_read_block(&freqProgram4, (unsigned char *)EEPROM_ADDRESS_PROGRAM4_FREQ, sizeof(freqProgram4));
	eeprom_read_block(&irmp_program5, (unsigned char *)EEPROM_ADDRESS_PROGRAM5, sizeof(irmp_program5));
	eeprom_read_block(&freqProgram5, (unsigned char *)EEPROM_ADDRESS_PROGRAM5_FREQ, sizeof(freqProgram5));
	eeprom_read_block(&irmp_program6, (unsigned char *)EEPROM_ADDRESS_PROGRAM6, sizeof(irmp_program6));
	eeprom_read_block(&freqProgram6, (unsigned char *)EEPROM_ADDRESS_PROGRAM6_FREQ, sizeof(freqProgram6));
	eeprom_read_block(&irmp_program7, (unsigned char *)EEPROM_ADDRESS_PROGRAM7, sizeof(irmp_program7));
	eeprom_read_block(&freqProgram7, (unsigned char *)EEPROM_ADDRESS_PROGRAM7_FREQ, sizeof(freqProgram7));
	eeprom_read_block(&irmp_program8, (unsigned char *)EEPROM_ADDRESS_PROGRAM8, sizeof(irmp_program8));
	eeprom_read_block(&freqProgram8, (unsigned char *)EEPROM_ADDRESS_PROGRAM8_FREQ, sizeof(freqProgram8));
	eeprom_read_block(&irmp_program9, (unsigned char *)EEPROM_ADDRESS_PROGRAM9, sizeof(irmp_program9));
	eeprom_read_block(&freqProgram9, (unsigned char *)EEPROM_ADDRESS_PROGRAM9_FREQ, sizeof(freqProgram9));
	eeprom_read_block(&irmp_rr, (unsigned char *)EEPROM_ADDRESS_RR, sizeof(irmp_rr));
	eeprom_read_block(&irmp_r, (unsigned char *)EEPROM_ADDRESS_R, sizeof(irmp_r));
	eeprom_read_block(&irmp_ff, (unsigned char *)EEPROM_ADDRESS_FF, sizeof(irmp_ff));
	eeprom_read_block(&irmp_f, (unsigned char *)EEPROM_ADDRESS_F, sizeof(irmp_f));
	eeprom_read_block(&irmp_mute, (unsigned char *)EEPROM_ADDRESS_MUTE, sizeof(irmp_mute));
	eeprom_read_block(&irmp_program, (unsigned char *)EEPROM_ADDRESS_PROGRAM, sizeof(irmp_program));

    for (;;)
    {
		wdt_reset(); //reset watchdog, we are still alive!
    
		cli();
		if (event & eventUsart0RxDInterrupt)
		{
			event &= ~eventUsart0RxDInterrupt;
			sei();
	    
			//collect the command
	    
			//switch group display on (G) or off (g)
			//if (usart0RxDByte == 'G')
			//displayRaw = 0;
			//else if (usart0RxDByte == 'g')
			//displayRaw = 1;
	    
			//reset decoder
			if (usart0RxDByte == 'r') {
				wdt_enable(WDTO_15MS);
				for(;;) {};
			}
	    
			//switch group display on (B) or off (b)
			//if (usart0RxDByte == 'B')
			//displayBitstream = 0;
			//else if (usart0RxDByte == 'b')
			//displayBitstream = 1;
	    
		} else {
			sei();
		}
    
		//we have got 26 bits raw rds data
		cli();
		if (event & eventBitstreamSnapshot) {
			event &= ~eventBitstreamSnapshot;
			sei();
	    
			if (displayBitstream) {
				printf_P(PSTR("B: 0x%07lX\r\n"), bitstreamData);
			}
		} else {
			sei();
		}
    
		cli();
		if (event & eventTimerOverflow) {
			event &= ~eventTimerOverflow;
			sei();
	    
			/* no rds signal message */
			if (syncQuality == 0 && syncMessage < 12) {
				syncMessage++;
				if (syncMessage == 12) {
					isRDS = 0;
					radioTextDisplayStart = 0;
					radioTextZeroDisplayCount = 0;
					radioTextEndDisplayCount = 0;
					printf_P(PSTR("INFO: 0x02, No RDS signal.\r\n"));
					programmeIdentificationCode = 0;
			    
					if (radioOn == 1)
					{
						showNoRds();
					}
				}
			}
	    
			if (syncQuality > 0)
			syncMessage = 0;
	    
			//if (PORTC & _BV(PC0))
			//PORTC &= ~_BV(PC0); //heartbeat led off
			//else
			//PORTC |= _BV(PC0); //heartbeat led on
	    
		} else {
			sei();
		}
    
		//reset vars when sync lost. Warning, sync lost happens 26 times as many as groupcomplete, do not do anything lenghthy here
		cli();
		if (event & eventSyncLost) {
			event &= ~eventSyncLost;
			sei();
			group0CodeBitsSteadyCount = 5;
		} else {
			sei();
		}
    
		cli();
		if (event & eventGroupComplete) {
			event &= ~eventGroupComplete;
			sei();
	    
			//Group type code
			groupType = block2 / 0x1000; //bits 12-15 (2^12)
	    
			//Group version code
			if (block2 & _BV(11))
			groupVersion = 'B';
			else
			groupVersion = 'A';
	    
			if (displayRaw)
			printf_P(PSTR("GROUP%02u%c: 0x%04X 0x%04X 0x%04X 0x%04X\r\n"), groupType, groupVersion, block1, block2, block3, block4);
	    
			//PI Codes of block 3 B-version groups are not decoded.
	    
			//Programme Identification code
			if (programmeIdentificationCode != block1) {
				programmeIdentificationCode = block1;
				printf_P(PSTR("PI: 0x%04X, Detected new station.\r\n"), programmeIdentificationCode);
				isRDS = 1;
		    		    
				/* reset variables because PI code changed */
				trafficProgrammeIdentificationCode = 0xFF;
				programmeTypeCode = 0xFF;
				trafficAnnouncementCode = 0xFF;
				musicSpeechSwitchScode = 0xFF;
		    
				group0CodeBitsSteadyCount = 0;
				for(i = 0; i < sizeof(programmeServiceName); i++)
				programmeServiceName[i] = 0xFF;
		    
				decoderIdentificationControlCode = 0xFF;
		    
				for(i = 0; i < sizeof(alternativeFrequencyCodes); i++)
				alternativeFrequencyCodes[i] = 0;
		    
				linkageActuator = 0xFF;
		    
				extendedCountryCode = 0;
		    
				for(i = 0; i < sizeof(radioText); i++) {
					radioText[i] = 0;
					radioTextPrevious[i] = 0;
				}
		    
				radioTextDisplayStart = 0;
				radioTextZeroDisplayCount = 0;
				radioTextEndDisplayCount = 0;
				radioTextMaxIndex = -1;
		    
				textSegmentAddress0Seen = 0;
		    
				textVersionPrevious = 0;
				textSegmentAddressPrevious = 0;
		    
				utcMinutesPrevious = 0xFF;
		    
			}
	    
			//Programme Type code
			if (programmeTypeCode != ((block2 / 0x20) & 0x1F)) {
				programmeTypeCode = ((block2 / 0x20) & 0x1F); //bits 5-9 (2^5)
				printf_P(PSTR("PTY: 0x%02X, "), programmeTypeCode);
				switch(programmeTypeCode) {
					case 0:
					printf_P(PSTR("None."));
					break;
					case 1:
					printf_P(PSTR("News."));
					break;
					case 2:
					printf_P(PSTR("Current Affairs."));
					break;
					case 3:
					printf_P(PSTR("Information."));
					break;
					case 4:
					printf_P(PSTR("Sport."));
					break;
					case 5:
					printf_P(PSTR("Education."));
					break;
					case 6:
					printf_P(PSTR("Drama."));
					break;
					case 7:
					printf_P(PSTR("Cultures."));
					break;
					case 8:
					printf_P(PSTR("Science."));
					break;
					case 9:
					printf_P(PSTR("Varied Speech."));
					break;
					case 10:
					printf_P(PSTR("Pop Music."));
					break;
					case 11:
					printf_P(PSTR("Rock Music."));
					break;
					case 12:
					printf_P(PSTR("Easy Listening."));
					break;
					case 13:
					printf_P(PSTR("Light Classics."));
					break;
					case 14:
					printf_P(PSTR("Serious Classics."));
					break;
					case 15:
					printf_P(PSTR("Other Music."));
					break;
					case 16:
					printf_P(PSTR("Weather."));
					break;
					case 17:
					printf_P(PSTR("Finance."));
					break;
					case 18:
					printf_P(PSTR("Children."));
					break;
					case 19:
					printf_P(PSTR("Social Affairs."));
					break;
					case 20:
					printf_P(PSTR("Religion."));
					break;
					case 21:
					printf_P(PSTR("Phone In."));
					break;
					case 22:
					printf_P(PSTR("Travel & Touring."));
					break;
					case 23:
					printf_P(PSTR("Leisure & Hobby."));
					break;
					case 24:
					printf_P(PSTR("Jazz Music."));
					break;
					case 25:
					printf_P(PSTR("Country Music."));
					break;
					case 26:
					printf_P(PSTR("National Music."));
					break;
					case 27:
					printf_P(PSTR("Oldies Music."));
					break;
					case 28:
					printf_P(PSTR("Folk Music."));
					break;
					case 29:
					printf_P(PSTR("Documentary."));
					break;
					case 30:
					printf_P(PSTR("Alarm Test."));
					break;
					case 31:
					printf_P(PSTR("Alarm - Alarm !"));
					break;
					default:
					printf_P(PSTR("Unknown."));
					break;
				}
				printf_P(PSTR("\r\n"));
			}
	    
			//Type 0 groups: Basic tuning and switching information
			if (groupType == 0) {
		    
				//Traffic Programme Identification code
				//Traffic announcement code
				if (trafficAnnouncementCode != ((block2 / 0x10) & 0x01) ||
				trafficProgrammeIdentificationCode != (block2 & _BV(10)) / 0x400) {
			    
					trafficAnnouncementCode = (block2 / 0x10) & 0x01; //bit 4 (2^4)
					trafficProgrammeIdentificationCode = (block2 & _BV(10)) / 0x400; //bit 10
			    
					printf_P(PSTR("TP&TA: 0x%02X 0x%02X, "), trafficProgrammeIdentificationCode, trafficAnnouncementCode);

					if (trafficProgrammeIdentificationCode == 0) {
				    
						if (trafficAnnouncementCode == 0) {
							printf_P(PSTR("No traffic announcements available."));
							} else {
							printf_P(PSTR("Traffic announcements available via EON on another station."));
						}
				    
					} else {
				    
						if (trafficAnnouncementCode == 0) {
							printf_P(PSTR("Traffic announcements available on this station and maybe via EON on another station."));
						} else {
							printf_P(PSTR("Traffic announcement in progress."));
						}
				    
					}
			    
					printf_P(PSTR("\r\n"));
			    
				}
		    
				//Music Speech switch code
				if (musicSpeechSwitchScode != ((block2 / 0x08) & 0x01)) {
					musicSpeechSwitchScode = (block2 / 0x08) & 0x01;  //bit 3 (2^3)
					printf_P(PSTR("MS: 0x%02X, "), musicSpeechSwitchScode);

					if (musicSpeechSwitchScode)
					printf_P(PSTR("Music is being broadcast or station does not use MS flag."));
					else
					printf_P(PSTR("Speech is being broadcast."));
					printf_P(PSTR("\r\n"));
				}
		    
				//Decode program service name and decoder identification control code
				group0CodeBits = block2 & 0x03; //0, 1, 2, 3;
		    
		    
				//TODO: improve decoderIdentificationControlCode detection, decouple from PS name
				//Decoder-identification control code-bit is bit 3 in block2
				if (group0CodeBits == 0) {
					if (block2 & 0x04)
					decoderIdentificationControlCodeNew |= _BV(3);
					else
					decoderIdentificationControlCodeNew &= ~_BV(3);
				}
				if (group0CodeBits == 1) {
					if (block2 & 0x04)
					decoderIdentificationControlCodeNew |= _BV(2);
					else
					decoderIdentificationControlCodeNew &= ~_BV(2);
				}
				if (group0CodeBits == 2) {
					if (block2 & 0x04)
					decoderIdentificationControlCodeNew |= _BV(1);
					else
					decoderIdentificationControlCodeNew &= ~_BV(1);
				}
				if (group0CodeBits == 3) {
					if (block2 & 0x04)
					decoderIdentificationControlCodeNew |= _BV(0);
					else
					decoderIdentificationControlCodeNew &= ~_BV(0);
				}
		    
				//fill in information
				i = group0CodeBits * 2;
				programmeServiceNameNew[i] = block4 / 0xFF; //bits 8-16 (2^8)
				programmeServiceNameNew[i + 1] = block4; //bit 0-8
		    
				if (programmeServiceNameNew[i] != programmeServiceName[i] ||
				programmeServiceNameNew[i + 1] != programmeServiceName[i + 1]) {
					//detected change, reset counter if not already counting
					if (group0CodeBitsSteadyCount > 4)
					group0CodeBitsSteadyCount = 0;
				}
		    
				//increase counter only when there are less then 4 received words
				if (group0CodeBitsSteadyCount < 4)
				group0CodeBitsSteadyCount++;
		    
				programmeServiceName[i] = programmeServiceNameNew[i];
				programmeServiceName[i + 1] = programmeServiceNameNew[i + 1];
		    
				//when we detected 4 new PS words then display it
				if (group0CodeBitsSteadyCount == 4) {
					printf_P(PSTR("PS: "));
					for(i = 0; i < sizeof(programmeServiceName) && i <= 7; i++) {
						printf_P(PSTR("%c"), programmeServiceName[i]);
					}
					printf_P(PSTR("\r\n"));
			    
					//prevent redisplay
					group0CodeBitsSteadyCount++;
			    
					if (decoderIdentificationControlCode != decoderIdentificationControlCodeNew) {
						decoderIdentificationControlCode = decoderIdentificationControlCodeNew;
						printf_P(PSTR("DI: 0x%02X"), decoderIdentificationControlCode);
				    
						if (decoderIdentificationControlCode & 0b0001)
						printf_P(PSTR(", Stereo"));
						else
						printf_P(PSTR(", Mono"));
				    
						if (decoderIdentificationControlCode & 0b0010)
						printf_P(PSTR(", Artificial Head"));
				    
						if (decoderIdentificationControlCode & 0b0100)
						printf_P(PSTR(", Compressed"));
				    
						if (decoderIdentificationControlCode & 0b1000)
						printf_P(PSTR(", Static PTY"));
						else
						printf_P(PSTR(", Dynamic PTY"));
				    
						printf_P(PSTR(".\r\n"));
					}
				}
		    
		    
				if (groupVersion == 'A') {
			    
					//Alternative frequency codes
					for (h = 0; h < 2; h++) {
						if (h == 0)
						j = block3; //first AF is in bits 0-7 of block3
						else
						j = block3 / 256; //second bits 8-15
				    
						//only frequencies we want, no control codes
						if (j >= 1 && j <= 204) {
							for(i = 0; i < sizeof(alternativeFrequencyCodes); i++) {
								if (alternativeFrequencyCodes[i] == j) {
									break;
								}
								if (alternativeFrequencyCodes[i] == 0) {
									alternativeFrequencyCodes[i] = j;
									//printf_P(PSTR("AF: 0x%02X, "), alternativeFrequencyCodes[i]);
							    
									m = 875 + j;
									//printf_P(PSTR("%u.%uMHz.\r\n"), m / 10, m % 10);
									break;
								}
							}
						}
				    
						if (j == 224 && alternativeFrequencyCodes[25] != j) {
							alternativeFrequencyCodes[25] = 224;
							printf_P(PSTR("AF: 0x%02X, This station has no alternate frequencies.\r\n"), j);
						}
					}
			    
					//TODO: LF/MF untested because lack of LF/MF station
					//Station transmits on LF/MF
					if (block3 == 250 && alternativeFrequencyCodes[26] != block3) {
						alternativeFrequencyCodes[26] = block3;
						j = block3 / 256;
				    
						printf_P(PSTR("AF: 0x%02X 0x%02X, "), block3, j);
				    
						//LF 153kHz-279kHz in 9 KHz steps
						if (j >= 1 && j <= 15) {
							m = 144 + (j * 9);
							printf_P(PSTR("%uKHz.\r\n"), m);
						} else
						//MF 531KHz-1602kHz in 9 KHz steps
						if (j >= 16 && j <= 135) {
							m = 387 + (j * 9);
							printf_P(PSTR("%uKHz.\r\n"), m);
						}
					}
				}
				//version B contains PI code in block3
		    
			} else
			//Type 1 groups: Programme Item Number and slow labeling codes
			if (groupType == 1) {
		    
				if (groupVersion == 'A') {
					//TODO: 5 bit Radio Paging Codes, bits 0-4 block2
			    
					//TODO: Add textual description of LA
					//Linkage Actuator
					if (linkageActuator != (block3 & _BV(15))) {
						linkageActuator = (block3 & _BV(15));
						printf_P(PSTR("LA: 0x%02X\r\n"), linkageActuator);
					}
			    
					//store variant code, bits 13-15
					i = (block3 / 0x2000) & 0x07;
			    
					//TODO: Paging
					if (i == 0) {
				    
						//TODO: Add textual description of ECC code
						//Extended Country Code, bits 0-7 block3
						if (extendedCountryCode != (block3 & 0xFF)) {
							extendedCountryCode = (block3 & 0xFF);
							printf_P(PSTR("ECC: 0x%02X\r\n"), extendedCountryCode);
						}
				    
				    
					} else
					//TODO: TMC identification
					if (i == 1) {
				    
					} else
					//TODO: Paging identification
					if (i == 2) {
				    
					} else
					//TODO: Language codes
					if (i == 3) {
				    
					} else
					//TODO: not assigned
					if (i == 4) {
				    
					} else
					//TODO: not assigned
					if (i == 5) {
				    
					} else
					//TODO: For use by broadcasters
					if (i == 6) {
				    
					} else
					//TODO: Identification of EWS channel
					if (i == 7) {
				    
					}
				}
		    
				//TODO: Programme Item Number, block 4
		    
			} else
			//Type 2 groups: RadioText
			if (groupType == 2) {
		    
				//text version A or B, bit 5 block 2
				if (block2 & 0x10)
				textVersion = 'B';
				else
				textVersion = 'A';
		    
				//block2 bit 0-3
				textSegmentAddress = (block2 & 0x0F);
		    
				//clean radioText when version changes
				if (textVersionPrevious != 0 &&
				textVersionPrevious != textVersion) {
			    
					for(i = 0; i < sizeof(radioText); i++) {
						radioText[i] = 0;
					}
			    
					textSegmentAddressPrevious = 0;
					textSegmentAddress0Seen = 0;
				}
		    
				//detected new start of text segment, normally address 0x00
				if (textSegmentAddressPrevious > textSegmentAddress) {
			    
					if (groupVersion == 'A')
					h = 64;
					else
					h = 32;
			    
					//detect new radioText
					j = 0;
					for (i = 0; i < h; i++) {
						if (radioText[i] != 0 && radioText[i] != ' ') {
							if (radioText[i] != radioTextPrevious[i]) {
								j = 1;
								break;
							}
						}
					}
			    
					//only print when we have received address 0 once.
					if (textSegmentAddress0Seen == 0)
					j = 0;
			    
					if (j) {
						printf_P(PSTR("RT%c: "), textVersion);

						for (i = 0; i < h; i++) {
					    
							if (radioText[i] == 0)
							break;
							else if (!(radioText[i] == '\r' || radioText[i] == '\n'))
							printf_P(PSTR("%c"), radioText[i]);

					    
							radioTextPrevious[i] = radioText[i];
						}
						printf_P(PSTR("\r\n"));
					}
			    
			    
				}
		    
				//64 bit messages in block 3 & 4
				if (groupVersion == 'A') {
			    
					radioText[(textSegmentAddress * 4)] = block3 / 256;
					radioText[(textSegmentAddress * 4) + 1] = block3;
					radioText[(textSegmentAddress * 4) + 2] = block4 / 256;
					radioText[(textSegmentAddress * 4) + 3] = block4;
			    
					//fill bytes smaller than textSegmentAddress with spaces when they are '0'
					if (textSegmentAddress > 0) {
						for (i = 0; i < (textSegmentAddress - 1) * 4; i++) {
							if (radioText[i] == 0)
							radioText[i] = ' ';
						}
					}
			    
				}
				//TODO: 32 bit messages not tested because of lack station transmitting it.
				//32 bit messages in block 4
				else {
					radioText[(textSegmentAddress * 2)] = block4 / 256;
					radioText[(textSegmentAddress * 2) + 1] = block4;
			    
					//fill bytes smaller then textSegmentAddress with spaces when they are '0'
					if (textSegmentAddress > 0) {
						for (i = 0; i < (textSegmentAddress - 1) * 2; i++) {
							if (radioText[i] == 0)
							radioText[i] = ' ';
						}
					}
			    
				}
		    
				if (textSegmentAddress == 0)
				textSegmentAddress0Seen = 1;
		    
				// determine radioTextMaxIndex
				radioTextMaxIndex = -1;
				// from end
				for (int q = 64; q >= 0; q--)
				{
					if (radioText[q] != ' ' && radioText[q] != '\r' && radioText[q] != '\n' && isprint(radioText[q]))
					{
						radioTextMaxIndex = q;
						break;
					}
				}
		    
				// from start
				for (int q = 0; q <= radioTextMaxIndex; q++)
				{
					if (radioText[q] < 16) // when radiotext changes from longer to shorter string, there may be remnants of old radiotext separated by some gibberish character, trim it
					{
						radioTextMaxIndex = (q - 1);
						break;
					}
				}
		    
				textVersionPrevious = textVersion;
				textSegmentAddressPrevious = textSegmentAddress;
		    
			} else
			//Type 3A groups: Application identification for Open data
			if (groupType == 3 && groupVersion == 'A') {
		    
			} else
			//Type 3B groups: Open Data Application
			if (groupType == 3 && groupVersion == 'B') {
		    
			} else
			//Type 4A groups : Clock-time and date
			if (groupType == 4 && groupVersion == 'A') {
		    
				//bits 0-5 are in block4 as bits 6-11
				utcMinutes = (block4  / 64) & 0x3F;
		    
				if (utcMinutesPrevious != utcMinutes) {
			    
					utcMinutesPrevious = utcMinutes;
			    
					//bits 0-14 are in block3 as bits 1-15
					//bits 15-16 are in block2 as bits 0-1
					modifiedJulianDay = (block3 / 2) + (block2 & 0x03) * 32768;
			    
					//bits 0-3 are in block4 as bits 12-15
					//bit 4 is in block3 as bit 1
					utcHours = (block4 / 4096) + (block3 & 0x01) * 16;
			    
					//local time offset are bits 0-4 in block 4
					localTimeOffset = block4 & 0x1F;
					//sign is in bit 5 of block4, 0=+ 1=-
					if (block4 & 0x20)
					localSign = '-';
					else
					localSign = '+';
			    
					//multiply by 30 so that we have offset in minutes (offset is in multiples of .5 hours)
					localTimeOffset *= 30;
			    
					printf_P(PSTR("CT: 0x%01X%04X%04X, "), (block2 & 0x03), block3, block4);
			    
					//Modified Julian date to year-month-day conversion
					utcYear = floor((modifiedJulianDay - 15078.2) / 365.25);
					utcMonth = floor((modifiedJulianDay - 14956.1 - floor(utcYear * 365.25)) / 30.6001);
					utcDay = modifiedJulianDay - 14956 - floor(utcYear * 365.25) - floor(utcMonth * 30.6001);
			    
					if (utcMonth == 14 || utcMonth == 15)
					i = 1;
					else
					i = 0;
			    
					utcYear = utcYear + i + 1900;
					utcMonth = utcMonth - 1 - (i * 12);
			    
					printf_P(PSTR("UTC %04u-%02u-%02u (MJD %u) %02u:%02u:00 %c%02u:%02u, "),
					utcYear, utcMonth, utcDay,
					modifiedJulianDay,
					utcHours, utcMinutes, localSign,
					localTimeOffset / 60, localTimeOffset % 60);
			    
					//TODO: half hour timezones and negative timezones not tested because lack of station transmitting it.
					//lets calulate local time
					if (localSign == '-') {
						localHours = utcHours - (localTimeOffset / 60);
						localMinutes = utcMinutes - (localTimeOffset % 60);
						} else {
						localHours = utcHours + (localTimeOffset / 60);
						localMinutes = utcMinutes + (localTimeOffset % 60);
					}
			    
					if (localMinutes < 0) {
						localMinutes += 60;
						localHours--;
					}
			    
					if (localMinutes > 59) {
						localMinutes -= 60;
						localHours++;
					}
			    
					if (localHours < 0)
					localHours += 24;
			    
					if (localHours > 23)
					localHours -= 24;
			    
					printf_P(PSTR("TIME %02u:%02u:00\r\n"), localHours, localMinutes);
				}

			} else
			//Type 4B groups: Open data application
			if (groupType == 4 && groupVersion == 'B') {
		    
			} else
			//Type 5 groups: Transparent data channels or ODA
			if (groupType == 5) {
		    
			} else
			//Type 6 groups: In-house applications or ODA
			if (groupType == 6) {
		    
			} else
			//Type 7A groups: Radio Paging or ODA
			if (groupType == 7 && groupVersion == 'A') {
		    
			} else
			//Type 7B groups: Open data application
			if (groupType == 7 && groupVersion == 'B') {
		    
			} else
			//Type 8 groups: Traffic Message Channel or ODA
			if (groupType == 8) {
		    
			} else
			//Type 9 groups: Emergency warning systems or ODA
			if (groupType == 9) {
		    
			} else
			//Type 10A groups: Programme Type Name
			if (groupType == 10 && groupVersion == 'A') {
		    
			} else
			//Type 10B groups: Open data
			if (groupType == 10 && groupVersion == 'A') {
		    
			} else
			//Type 11 groups: Open Data Application
			if (groupType == 11) {
		    
			} else
			//Type 12 groups: Open Data Application
			if (groupType == 12) {
		    
			} else
			//Type 13A groups: Enhanced Radio Paging or ODA
			if (groupType == 13 && groupVersion == 'A') {
		    
			} else
			//Type 13B groups: Open Data Application
			if (groupType == 13 && groupVersion == 'B') {
		    
			} else
			//Type 14 groups: Enhanced Other Networks information
			if (groupType == 14) {
				printf_P(PSTR("GROUP%02u%c: 0x%04X 0x%04X 0x%04X 0x%04X\r\n"), groupType, groupVersion, block1, block2, block3, block4);
			} else
			//Type 15A groups: 'currently unavailable'
			if (groupType == 15 && groupVersion == 'A') {
		    
			} else
			//Type 15B groups: Fast basic tuning and switching information
			if (groupType == 15 && groupVersion == 'B') {
		    
			}
		} else {
			sei();
		}
	}
}

